import numpy as np


class Sigmoid:
    def __init__(self):
        """A logistic sigmoid activation function."""

    def __str__(self):
        return "Sigmoid"

    @staticmethod
    def fn(z):
        return 1 / (1 + np.exp(-z))

    def grad(self, x):
        fn_x = self.fn(x)
        return fn_x * (1 - fn_x)


class ReLU:
    def __init__(self):
        """A rectified linear activation function."""

    def __str__(self):
        return "ReLU"

    @staticmethod
    def fn(z):
        return np.clip(z, 0, np.inf)

    @staticmethod
    def grad(x):
        return (x > 0).astype(int)

    @staticmethod
    def grad2(x):
        return np.zeros_like(x)


class LeakyReLU:
    """
    'Leaky' version of a rectified linear unit (ReLU).
    Notes
    -----
    Leaky ReLUs [*]_ are designed to address the vanishing gradient problem in
    ReLUs by allowing a small non-zero gradient when `x` is negative.
    Parameters
    ----------
    alpha: float
        Activation slope when x < 0. Default is 0.3.
    """

    def __init__(self, alpha=0.3):
        self.alpha = alpha

    def __str__(self):
        """Return a string representation of the activation function"""
        return "Leaky ReLU(alpha={})".format(self.alpha)

    def fn(self, z):
        _z = z.copy()
        _z[z < 0] = _z[z < 0] * self.alpha
        return _z

    def grad(self, x):
        out = np.ones_like(x)
        out[x < 0] *= self.alpha
        return out

    @staticmethod
    def grad2(x):
        return np.zeros_like(x)


class Tanh:
    def __init__(self):
        """A hyperbolic tangent activation function."""

    def __str__(self):
        """Return a string representation of the activation function"""
        return "Tanh"

    @staticmethod
    def fn(z):
        """Compute the tanh function on the elements of input `z`."""
        return np.tanh(z)

    @staticmethod
    def grad(x):
        return 1 - np.tanh(x) ** 2

    @staticmethod
    def grad2(x):
        tanh_x = np.tanh(x)
        return -2 * tanh_x * (1 - tanh_x**2)


class Affine:
    def __init__(self, slope=1, intercept=0):
        """
        An affine activation function.
        Parameters
        ----------
        slope: float
            Activation slope. Default is 1.
        intercept: float
            Intercept/offset term. Default is 0.
        """
        self.slope = slope
        self.intercept = intercept

    def __str__(self):
        """Return a string representation of the activation function"""
        return "Affine(slope={}, intercept={})".format(self.slope, self.intercept)

    def fn(self, z):
        return self.slope * z + self.intercept

    def grad(self, x):
        return self.slope * np.ones_like(x)

    @staticmethod
    def grad2(x):
        return np.zeros_like(x)


class Identity(Affine):
    def __init__(self, *args, **kwargs):
        """
        Identity activation function.
        Notes
        -----
        :class:`Identity` is just syntactic sugar for :class:`Affine` with
        slope = 1 and intercept = 0.
        """
        super().__init__(*args, **kwargs)

    def __str__(self):
        return "Identity"


class ELU:
    def __init__(self, alpha=1.0):
        r"""
        An exponential linear unit (ELU).

        alpha : float
            Slope of negative segment. Default is 1.
        """
        self.alpha = alpha

    def __str__(self):
        return "ELU(alpha={})".format(self.alpha)

    def fn(self, z):
        """z if z > 0  else alpha * (e^z - 1)"""
        return np.where(z > 0, z, self.alpha * (np.exp(z) - 1))

    def grad(self, x):
        # 1 if x > 0 else alpha * e^(z)
        return np.where(x > 0, np.ones_like(x), self.alpha * np.exp(x))

    def grad2(self, x):
        # 0 if x > 0 else alpha * e^(z)
        return np.where(x >= 0, np.zeros_like(x), self.alpha * np.exp(x))


class Exponential:
    def __init__(self):
        """An exponential (base e) activation function"""

    def __str__(self):
        return "Exponential"

    @staticmethod
    def fn(z):
        return np.exp(z)

    @staticmethod
    def grad(x):
        return np.exp(x)

    @staticmethod
    def grad2(x):
        return np.exp(x)


class SELU:
    r"""
    A scaled exponential linear unit (SELU).
    """

    def __init__(self):
        self.alpha = 1.6732632423543772848170429916717
        self.scale = 1.0507009873554804934193349852946
        self.elu = ELU(alpha=self.alpha)

    def __str__(self):
        return "SELU"

    def fn(self, z):
        return self.scale * self.elu.fn(z)

    def grad(self, x):
        return np.where(
            x >= 0,
            np.ones_like(x) * self.scale,
            np.exp(x) * self.alpha * self.scale,
        )

    def grad2(self, x):
        return np.where(x > 0, np.zeros_like(x), np.exp(x) * self.alpha * self.scale)


class HardSigmoid:
    def __init__(self):
        """
        Notes
        -----
        The hard sigmoid is a piecewise linear approximation of the logistic
        sigmoid that is computationally more efficient to compute.
        """

    def __str__(self):
        return "Hard Sigmoid"

    @staticmethod
    def fn(z):
        return np.clip((0.2 * z) + 0.5, 0.0, 1.0)

    @staticmethod
    def grad(x):
        return np.where((x >= -2.5) & (x <= 2.5), 0.2, 0)

    @staticmethod
    def grad2(x):
        return np.zeros_like(x)


class SoftPlus:
    def __init__(self):
        """
        A softplus activation function.
        Notes
        -----
        In contrast to :class:`ReLU`, the softplus activation is differentiable
        everywhere (including 0). It is, however, less computationally efficient to
        compute.
        The derivative of the softplus activation is the logistic sigmoid.
        """

    def __str__(self):
        return "SoftPlus"

    @staticmethod
    def fn(z):
        return np.log(np.exp(z) + 1)

    @staticmethod
    def grad(x):
        exp_x = np.exp(x)
        return exp_x / (exp_x + 1)

    @staticmethod
    def grad2(x):
        exp_x = np.exp(x)
        return exp_x / ((exp_x + 1) ** 2)



"""
if __name__ == "__main__":
    import uvicorn
    from {{project_name}}.api import create_application

    app = create_application()
    uvicorn.run(
        app,
        host="0.0.0.0",
        port=8080,
        app_dir=".",
        reload=False,
        log_config="conf/uvicorn.log.yaml",
    )
"""



from .core import *
from .ops import *
from subprocess import Popen, PIPE
from pathlib import Path


def run(cmd, **env):
    cmd = cmd.split(" ") if isinstance(cmd, str) else cmd
    p = Popen(cmd, cwd=str(Path(__file__).parent), env={**os.environ, **env})
    p.communicate()
    return p.returncode



import numpy as np
from .utils import is_binary, is_stochastic


class SquaredError:
    def __init__(self):
        """
        A squared-error or known as  `L2` loss.
        """

    def __call__(self, y, y_pred):
        return self.loss(y, y_pred)

    def __str__(self):
        return "SquaredError"

    @staticmethod
    def loss(y, y_pred):
        return 0.5 * np.linalg.norm(y_pred - y) ** 2

    @staticmethod
    def grad(y, y_pred, z, act_fn):
        return (y_pred - y) * act_fn.grad(z)


class CrossEntropy:
    def __init__(self):
        """
        A cross-entropy loss.
        """

    def __call__(self, y, y_pred):
        return self.loss(y, y_pred)

    def __str__(self):
        return "CrossEntropy"

    @staticmethod
    def loss(y, y_pred):
        is_binary(y)
        is_stochastic(y_pred)

        # prevent taking the log of 0
        eps = np.finfo(float).eps

        # each example is associated with a single class; sum the negative log
        # probability of the correct label over all samples in the batch.
        # observe that we are taking advantage of the fact that y is one-hot
        # encoded
        cross_entropy = -np.sum(y * np.log(y_pred + eps))
        return cross_entropy

    @staticmethod
    def grad(y, y_pred):
        is_binary(y)
        is_stochastic(y_pred)

        # derivative of xe wrt z is y_pred - y_true, hence we can just
        # subtract 1 from the probability of the correct class labels
        grad = y_pred - y

        # [optional] scale the gradients by the number of examples in the batch
        # n, m = y.shape
        # grad /= n
        return grad



import os
from typing import Tuple, List, NamedTuple
import psutil
import subprocess


class ProcessInfo(NamedTuple):
    port: int
    process: psutil.Process


def get_pid(port):
    port = int(port)
    cmd = 'lsof -t -i:{0}'.format(port)
    pid = None
    try:
        pid = subprocess.check_output(cmd, shell=True)
    except Exception as e:
        print("No process running on port {} by current user. Checking if root is running the proecess".format(port))
        if pid is None:
            cmd = 'sudo lsof -t -i:{0}'.format(port)
            pid = subprocess.check_output(cmd, shell=True)
    pids = pid.decode().split("\n")
    pids_int = []
    for pid in pids:
        if pid:
            pid = int(pid)
            pids_int.append(pid)
    return pids_int


def kill_port_by_lsof(port: int):
    pids = get_pid(port)
    for pid in pids:
        try:
            os.kill(pid, 9)
            emoji = "☠️🔪"
            print(
                f"{emoji}: pid {pid} on port {port}",
            )
        except Exception as e:
            print(
                f"kill: pid {pid} failed"
            )



def get_processes(ports: List[int]) -> List[ProcessInfo]:
    processes = set()
    for process in psutil.process_iter():
        try:
            conns = process.connections(kind="inet")
        except (psutil.AccessDenied, psutil.ZombieProcess):
            continue

        for conn in conns:
            port = conn.laddr.port
            if port in ports:
                processes.add(ProcessInfo(port=port, process=process))

    return sorted(processes, key=lambda p: p.port)


def kill_ports(ports: Tuple[int], just_view: bool = False) -> int:
    processes = get_processes(list(ports))
    if not processes:
        print(f"🙃 No processes found for the given port: {ports}")
        return False

    for pinfo in processes:
        emoji = "👁️" if just_view else "☠️🔪"
        process = pinfo.process
        if not just_view:
            process.kill()
        print(
            f"{emoji}: {process.name()} (pid {process.pid}) "
            f"on port {pinfo.port}",
        )
    return True if processes else False



import numbers
import numpy as np


def is_stochastic(X):
    """True if `X` contains probabilities that sum to 1 along the columns"""
    msg = "Array should be stochastic along the columns"
    assert len(X[X < 0]) == len(X[X > 1]) == 0, msg
    assert np.allclose(np.sum(X, axis=1), np.ones(X.shape[0])), msg
    return True


def is_number(a):
    """Check that a value `a` is numeric"""
    return isinstance(a, numbers.Number)


def is_one_hot(x):
    """Return True if array `x` is a binary array with a single 1"""
    msg = "Matrix should be one-hot binary"
    assert np.array_equal(x, x.astype(bool)), msg
    assert np.allclose(np.sum(x, axis=1), np.ones(x.shape[0])), msg
    return True


def is_binary(x):
    """Return True if array `x` consists only of binary values"""
    msg = "Matrix must be binary"
    assert np.array_equal(x, x.astype(bool)), msg
    return True



import subprocess
import io
import csv
import collections


def commandexists(shellcommand):
    status, output = subprocess.getstatusoutput(shellcommand)
    exists = status == 0
    if not exists:
        print("Could not execute: {0}".format(shellcommand))
    return exists


def command(args):
    return subprocess.check_output(args).decode()


def csvtodictdict(csvdata, colnames, keycols, fmtcols={}):
    """
    Returns a dict of dicts from csv file with specified column names and primary key column
    accepts and optional element formatting per column as a dictionary of format functions
    """
    fmtcols = collections.defaultdict(lambda: lambda x: x, **fmtcols)
    d = {}
    rows = csv.reader(csvdata)
    for row in rows:
        drow = {colname: fmtcols[colname](val) for colname, val in zip(colnames, row)}
        if isinstance(keycols, str):
            key = drow.pop(keycols)
        else:
            key = tuple([drow.pop(keycol) for keycol in keycols])
        d[key] = drow
    return d


def csvheaderargs(fmtcol, cols):
    return ",".join([fmtcol.format(col) for col in cols])


def commandtodictdict(baseargs, cols, keycols=None, queryargfmt="{0}", colargfmt="{0}", outputfmt={}, skipheader=False):
    queryarg = queryargfmt.format(csvheaderargs(colargfmt, cols))
    args = baseargs + [queryarg]
    csvoutput = io.StringIO(command(args))
    if skipheader:
        csvoutput.readline()
    if keycols is None:
        keycols = cols[0]
    return csvtodictdict(csvoutput, cols, keycols, fmtcols=outputfmt)


def renamekeys(d, names):
    '''
    updates key names in d based on dict of old/new name pairs
    returning resulting updated dict
    '''
    for oldname, newname in names.items():
        d[newname] = d.pop(oldname)
    return d


def docker_gpu_stat():
    # get results of all commands without container arguments
    dockerps = commandtodictdict(['docker', 'ps', '--format'],
                                 ['ID', 'Image', 'Ports'],
                                 keycols='ID',
                                 queryargfmt="'{0}'",
                                 colargfmt="{{{{.{0}}}}}",
                                 outputfmt={'ID': lambda s: s[1:]})
    dockerstats = commandtodictdict(['docker', 'stats', '--no-stream', '--format'],
                                    ['Container', 'MemUsage', 'CPUPerc'],
                                    keycols='Container',
                                    queryargfmt="'{0}'",
                                    colargfmt="{{{{.{0}}}}}",
                                    outputfmt={'Container': lambda s: s[1:]})
    unitstats = commandtodictdict(['nvidia-smi', '--format=csv'],
                                  ['gpu_uuid', 'utilization.gpu', 'utilization.memory'],
                                  keycols='gpu_uuid',
                                  queryargfmt="--query-gpu={0}",
                                  outputfmt={'gpu_uuid': lambda s: s.lstrip()},
                                  skipheader=True)
    unitprocstats = commandtodictdict(['nvidia-smi', '--format=csv'],
                                      ['pid', 'process_name', 'gpu_uuid', 'used_memory'],
                                      keycols=['pid', 'gpu_uuid'],
                                      queryargfmt="--query-compute-apps={0}",
                                      outputfmt={'gpu_uuid': lambda s: s.lstrip()},
                                      skipheader=True)

    # map gpu_uuids to short ids in unit info rename columns
    shortunitids = {gpu_uuid: "{0}".format(shortid) for gpu_uuid, shortid in
                    zip(unitstats.keys(), range(len(unitstats)))}
    colnames = {'utilization.gpu': 'used_gpu'}
    unitstats = {shortunitids[gpu_uuid]: renamekeys(stats, colnames) for gpu_uuid, stats in unitstats.items()}
    unitprocstats = {(pid, shortunitids[gpu_uuid]): stats for (pid, gpu_uuid), stats in unitprocstats.items()}

    # display fmt data
    basedisplaycols = collections.OrderedDict([('Container', 12),
                                               ('Image', 18)])
    optdisplaycols = collections.OrderedDict([('pid', 7),
                                              ('gpu_uuid', 8),
                                              ('used_memory', 12),
                                              ('used_gpu', 9)])
    displaycols = collections.OrderedDict(list(basedisplaycols.items()) +
                                          list(optdisplaycols.items()))

    # display fmt strings
    basedisplayfmt = '\t'.join(['{{{0}:{1}.{1}}}'.format(col, width) for col, width in basedisplaycols.items()])
    optdisplayfmt = '\t'.join(['{{{0}:{1}.{1}}}'.format(col, width) for col, width in optdisplaycols.items()])
    displayfmt = '\t'.join([basedisplayfmt, optdisplayfmt])

    # print rows of relevant container processes
    # (everything below a bit janky in terms of argument expectations and generalization)
    dockerall = {container: {**dockerps[container], **dockerstats[container]} for container in dockerstats.keys()}
    someunitsactive = False
    print(displayfmt.format(**{col: col for col in displaycols.keys()}))
    for container, dockerinfo in dockerall.items():
        # very particular incantation needed here for top options to function correctly:
        # https://www.projectatomic.io/blog/2016/01/understanding-docker-top-and-ps/
        pids = command(['docker', 'top', container, '-eo', 'pid']).split('\n')[1:-1]  # obviously could be a bit brittle

        containerunitstatslist = [((proc, unit), stats) for (proc, unit), stats in sorted(unitprocstats.items()) if
                                  proc in pids]
        containerunitstats = collections.OrderedDict(containerunitstatslist)

        if containerunitstats:
            someunitsactive = True
            basedisplaystr = basedisplayfmt.format(Container=container, **dockerinfo)
            print(basedisplaystr)
            for (pid, gpu_uuid), stats in containerunitstats.items():
                print(optdisplayfmt.rjust(99).format(pid=pid, gpu_uuid=gpu_uuid, **stats, **unitstats[gpu_uuid]))
    if not someunitsactive:
        print("\n\t\t no gpu units being used by docker containers ")


if __name__ == '__main__':
    # check for existence of docker and nvidia-smi commands
    if commandexists('docker') and commandexists('nvidia-smi'):
        docker_gpu_stat()
    else:
        print('Command(s) not found')



import base64
import numpy as np
import cv2


def base64_to_numpy(image_base64):
    image_bytes = base64.b64decode(image_base64)
    image_np = np.frombuffer(image_bytes, dtype=np.uint8)
    image_np2 = cv2.imdecode(image_np, cv2.IMREAD_COLOR)
    # image_np2 = cv2.cvtColor(image_np2, cv2.COLOR_BGR2RGB)
    image_np2 = image_np2 / 255
    return image_np2.astype(np.float32)


def numpy_to_base64(image, unnormalize: bool):
    if unnormalize:
        image = image / 2 + 0.5  # unnormalize
    if image.dtype in (np.float32, np.double) and image.max() <= 1:
        image = (image * 255).astype('uint8')
    retval, buffer = cv2.imencode('.jpg', image)
    pic_str = base64.b64encode(buffer)
    return pic_str.decode()


def numpy_to_tensor(image_np, normalize=True):
    if normalize:
        image_np = (image_np - 0.5) * 2
    return np.transpose(image_np, (2, 0, 1))[None, ...]


def tensor_to_base64(image_tensor, unnormalize=True):
    import torch
    if image_tensor.dtype == torch.float:
        image_tensor = np.array(image_tensor.cpu())
    b64_list = [numpy_to_base64(np.transpose(image, (1, 2, 0)), unnormalize) for image in image_tensor]
    return b64_list


def base64_list_to_tensor(b64_list, normalize=True):
    tensor_list = [numpy_to_tensor(base64_to_numpy(b64), normalize) for b64 in b64_list]
    return np.concatenate(tensor_list, 0)



from __future__ import annotations
from typing import List, Dict
from abc import ABCMeta, abstractmethod
from pathlib import Path


class Trie(metaclass=ABCMeta):
    trie = None
    rtrie = None

    def matches(self, word: str) -> list:
        matched_list = self.startwith(word)
        matched_list.extend(self.prefixes(word))
        return matched_list

    def __contains__(self, item):
        return item in self.trie

    @abstractmethod
    def prefixes(self, word: str) -> list:
        """在trie中匹配出所有可以作为word的前缀的keys

        Parameters
        ----------
        word : str
            待匹配词语
            
        Returns
        -------
        list
            Trie中所有匹配到的元素数组

        Example
        -------
        >>> trie = Trie(["ab", "bc", "bcd", "bcde"])
        >>> trie.prefixes("bcdf")
        ["bc", "bcd"]
        """

    def full_prefixes(self, word: str) -> list:
        full_matches = []
        for idx in range(len(word)):
            cur_word = word[idx:]
            lst = self.prefixes(cur_word)
            lst.sort(key=lambda x: len(x), reverse=True)
            full_matches.extend(lst)
        return full_matches

    @abstractmethod
    def startwith(self, word: str) -> list:
        """在前缀树中匹配出所有以word为前缀的keys

        Parameters
        ----------
        word : str
            待匹配词语
            
        Returns
        -------
        list
            Trie中所有匹配到的元素数组
        
        Example
        -------
        >>> trie = Trie(["ab", "bc", "bcd", "bcde"])
        >>> trie.startwith("bc")
        ["bc", "bcd", "bcde"]
        """

    @abstractmethod
    def endwith(self, word: str) -> list:
        """在前缀树中匹配出所有以word为后缀的keys

        Parameters
        ----------
        word : str
            待匹配词语

        Returns
        -------
        list
            Trie中所有匹配到的元素数组

        Example
        -------
        >>> trie = Trie(["bc", "abc", "bcd", "edbc"])
        >>> trie.endwith("bc")
        ["bc", "abc", "edbc"]
        """

    def has_keys_with_prefix(self, word):
        ...

    def save(self):
        ...

    def load(self, file_name: str | Path):
        ...


class PyTrie:
    def __init__(self, words: List[str]):
        self.trie = {}
        self._end = "eos"
        for word in words:
            self.add(word)

    def add(self, word):
        word = word.strip()
        assert word
        cur_node = self.trie
        for char in word:
            cur_node = cur_node.setdefault(char, {})
        # cur_node[self._end] = ""
        cur_node[self._end] = word

    def find(self, word: str, full=True) -> bool:
        """
        Parameters
        ----------
        full: bool
            True: 完全匹配时返回True
            False: 前缀匹配到则返回True
        """
        trie = self.trie
        for c in word:
            if c not in trie:
                return False
            trie = trie[c]
        if full:
            return self._end in trie
        else:
            return True

    def extract_longest_item(self, word: str):
        """从一个文本的开头开始匹配字典中最长的词，返回最长词和长度"""
        curr_dict, longest, offset = self.trie, None, 0

        if not word:
            return longest, offset

        for i, c in enumerate(word):
            if c not in curr_dict:
                return longest, offset
            curr_dict = curr_dict[c]
            if 'end' in curr_dict:
                longest, offset = curr_dict['end'], i + 1
        return longest, offset

    def __str__(self):
        return self.trie


class HatTrie(Trie):
    ...


class MarisaTrie(Trie):
    """Static memory-efficient Trie-like structures for Python
    based on marisa-trie C++ library.
    String data in a MARISA-trie may take up to 50x-100x less memory than in a standard Python dict;
    the raw lookup speed is comparable;
    trie also provides fast advanced methods like prefix search.
    """

    def __init__(self, key_list: list):
        import marisa_trie
        self.trie = marisa_trie.Trie(key_list)
        rkey_list = [i[::-1] for i in key_list]
        self.rtrie = marisa_trie.Trie(rkey_list)

    def prefixes(self, word: str):
        return self.trie.prefixes(word)

    def rprefixes(self, word: str):
        rword = word[::-1]
        return self.rtrie.prefixes(rword)

    def startwith(self, word: str):
        return self.trie.keys(word)

    def endwith(self, word: str):
        rword = word[::-1]
        return [i[::-1] for i in self.rtrie.keys(rword)]

    def get_longest(self, text):
        """匹配整个字典中最长的词，返回最长词"""
        if not text:
            return None

        longest = ''
        for idx, item in enumerate(text):
            items = self.trie.prefixes(text[idx:])

            for item in items:
                if len(item) > len(longest):
                    longest = item
        if longest == '':
            return None
        return longest

    def has_keys_with_prefix(self, word):
        return self.trie.has_keys_with_prefix(word)


class DaTrie(Trie):
    def __init__(self, word_list: list):
        import datrie
        self.trie = datrie.BaseTrie(word_list)
        rkey_list = [i[::-1] for i in word_list]
        self.rtrie = datrie.BaseTrie(rkey_list)

    def prefixes(self, name: str):
        return self.trie.prefixes(name)

    def startwith(self, name: str):
        return self.trie.startwith(name)

    def longest_prefix(self, word: str):
        return self.trie.longest_prefix(word)

    def has_keys_with_prefix(self, word):
        return self.trie.has_keys_with_prefix(word)


class AutomatonTrie:
    def __init__(self, key_list: list):
        import ahocorasick
        self.trie = ahocorasick.Automaton()
        for idx, key in enumerate(key_list):
            self.trie.add_word(key, (idx, key))

    def __contains__(self, item):
        return item in self.trie


class Benchmark:
    def __init__(self):
        from sparrow.performance import MeasureTime
        self.ms = MeasureTime()
        n = 2000000
        word_list = [str(i) for i in range(n)]
        self.ms.start()
        pytrie = PyTrie(word_list)
        self.ms.show_interval(f"pytrie build")
        mtrie = MarisaTrie(word_list)
        self.ms.show_interval(f"marisa trie build")

    def run(self):
        self.ms.start()


if __name__ == "__main__":
    from rich import print

    key_list = ['as', 'asdf', "basdfg", 'casd']
    t = MarisaTrie(key_list)
    print(t.startwith("as"))
    print(t.endwith("sd"))
    print('asd' in t)
    t2 = AutomatonTrie(key_list)
    print('asd' in t2)

    # print(t.rkeys("asd"))
    # t = AutomatonTrie(['as', '1asdf'])
    # print("a" in t.trie)
    # t = PyTrie(["a", "ab", "abc", "bc",
    #             "abcd"
    #             ])
    # print(t.find("abc"))
    # print(t.find("bcd", ))
    # Benchmark()



import time
# from ..string.color_string import rgb_string, color_const
from rich.text import Text
from rich import print

__all__ = ['MeasureTime']


class MeasureTime:
    def __init__(self, prec=2, logger=None, gpu=False):
        if gpu:
            import torch
            self.cuda_sync = torch.cuda.synchronize
        self._cost_time = 0.
        self._start_time = time.time()
        self._logger = logger
        self._msg = None
        self._prec: int = prec

    def start(self, gpu=False):
        if gpu:
            self.cuda_sync()
        self._start_time = time.time()
        return self

    def show_interval(self, msg=None, gpu=False):
        if gpu:
            self.cuda_sync()
        self._msg = msg
        self._cost_time = time.time() - self._start_time
        self._start_time = time.time()
        self._show()
        return self.get_cost()

    def get_cost(self):
        return f"{self._cost_time:.{int(self._prec)}E}"

    def _show(self):
        cost_time = self.get_cost()
        msg = f"{self._msg}\t" if self._msg else ''
        if self._logger:
            show_string = f"{msg}cost time: {cost_time}s"
            self._logger.debug(show_string)
        else:
            rgb_cost_time = Text(cost_time, style='green')
            rgb_msg = Text(f"{msg}", style="cyan")
            str_tuple = (rgb_msg, 'cost time: ', rgb_cost_time)
            print(*str_tuple, sep='')



import re
from itertools import groupby


class Extractor():

    def extract_email(self, text):
        if text=='':
            return []
        eng_texts = self.remove_chinese(text)
        eng_texts = eng_texts.replace(' at ','@').replace(' dot ','.')
        sep = ',!?:; ，。！？《》、|\\/'
        eng_split_texts = [''.join(g) for k, g in groupby(eng_texts, sep.__contains__) if not k]

        email_pattern = r'^[a-zA-Z0-9_-]+@[a-zA-Z0-9_-]+(\.[a-zA-Z_-]+)+$'

        emails = []
        for eng_text in eng_split_texts:
            result = re.match(email_pattern, eng_text, flags=0)
            if result:
                emails.append(result.string)
        return emails


    def remove_chinese(self, text):
        """
        remove all the chinese characters in text
        """
        if text=='':
            return []
        filtrate = re.compile(u'[\u4E00-\u9FA5]')
        text_without_chinese = filtrate.sub(r' ', text)
        return text_without_chinese

    def extract_cellphone(self, text, nation="CH"):
        if text=='':
            return []
        eng_texts = self.remove_chinese(text)
        sep = ',!?:; ：，.。！？《》、|\\/abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
        eng_split_texts = [''.join(g) for k, g in groupby(eng_texts, sep.__contains__) if not k]
        eng_split_texts_clean = [ele for ele in eng_split_texts if len(ele)>=7 and len(ele)<17]

        if nation=='CH':
        #     phone_pattern = r'^((\+86)?([- ])?)?(|(13[0-9])|(14[0-9])|(15[0-9])|(17[0-9])|(18[0-9])|(19[0-9]))([- ])?\d{3}([- ])?\d{4}([- ])?\d{4}$'
            phone_pattern = r"^(?:\+?86)?1(?:3\d{3}|5[^4\D]\d{2}|8\d{3}|7(?:[0-35-9]\d{2}|4(?:0\d|1[0-2]|9\d))|9[0-35-9]\d{2}|6[2567]\d{2}|4(?:(?:10|4[01])\d{3}|[68]\d{4}|[579]\d{2}))\d{6}$"

        phones = []
        for eng_text in eng_split_texts_clean:
            result = re.match(phone_pattern, eng_text, flags=0)
            if result:
                phones.append((result.string.replace('+86','').replace('-',''), ))
        return phones
    
if __name__ == "__main__":
    print(Extractor().extract_cellphone("我的手机号是13121572916"))




#
#   Hello World server in Python
#   Binds REP socket to tcp://*:5555
#   Expects b"Hello" from client, replies with b"World"
#

import time
import zmq

context = zmq.Context()
socket = context.socket(zmq.REP)
socket.bind("tcp://*:5555")

while True:
    #  Wait for next request from client
    message = socket.recv()
    print("Received request: %s" % message)

    #  Do some 'work'
    # time.sleep(0.0001)

    #  Send reply back to client
    socket.send(b"World")


import pysnooper
import random


if __name__ == "__main__":
    def foo():
        lst = []
        for i in range(10):
            lst.append(random.randrange(1, 1000))

        with pysnooper.snoop():
            lower = min(lst)
            upper = max(lst)
            mid = (lower + upper) / 2
            print(lower, mid, upper)

    foo()



import pynvml
import time


class GpuStat:
    def __init__(self):
        pynvml.nvmlInit()

    def stop(self):
        pynvml.nvmlShutdown()

    def start_stat(self, interval=1, log=None):
        # 在线程中执行
        while True:
            time.sleep(interval)
            if log:
                log.debug(self.gpu_stat())
            else:
                print(self.gpu_stat())

    def gpu_stat(self):
        # https://pypi.org/project/nvidia-ml-py/
        gpu_num = pynvml.nvmlDeviceGetCount()
        gpu_version = pynvml.nvmlSystemGetDriverVersion()
        result = {}
        for i in range(gpu_num):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            info = pynvml.nvmlDeviceGetMemoryInfo(handle)
            total_memory = info.total / 1024 ** 2
            used_memory = info.used / 1024 ** 2
            free_memory = total_memory - used_memory
            # print(f"Device {i} : {pynvml.nvmlDeviceGetName(handle)}")
            # print(f"gpu num: {i}, total: {total_memory}Mb, used: {used_memory}mb free: {free_memory}Mb \n")
            # 获取显卡的工作模式，0 为 WDDM 模式、1 为 TCC 模式
            # WDDM 模式：GPU 负责计算以及图像显示，打游戏显然是 WDDM 模式
            # TCC 模式：GPU 只负责计算，不负责图像显示，机器学习的时候一般是 TCC 模式
            # print(pynvml.nvmlDeviceGetDriverModel(handle))  # [0, 0]
            """
            # 可以设置显卡工作模式，如果报如下错误
            # pynvml.NVMLError_NoPermission: Insufficient Permissions
            # 那么以管理员的身份打开命令行，然后执行
            pynvml.nvmlDeviceSetDriverModel(handle, 1)
            # 当然，如果显卡不是 Tesla 和 Quadro，那么不支持设置工作模式，此时会抛出如下错误：
            # pynvml.NVMLError_NotSupported: Not Supported
            """
            # fans = pynvml.nvmlDeviceGetFanSpeed(handle)
            temperature = pynvml.nvmlDeviceGetTemperature(handle, i)
            powerstate = pynvml.nvmlDeviceGetPowerState(handle)
            # gpu计算核心使用率
            gpu_rate = pynvml.nvmlDeviceGetUtilizationRates(handle).gpu
            # gpu内存使用率
            memory_rate = pynvml.nvmlDeviceGetUtilizationRates(handle).memory

            pid_info_list = pynvml.nvmlDeviceGetComputeRunningProcesses(handle)
            # print(temperature, powerstate, gpu_rate, memory_rate)
            for pid_info in pid_info_list:
                # pid_info.pid: 此进程的pid，再基于 psutil.Process 可以得到更详细的信息
                # pid_info.usedGpuMemory: 此进程使用的显存大小
                ...

            info = {"gpu_rate": gpu_rate,
                    "temperature": temperature,
                    "used_memory": used_memory,
                    "free_memory": free_memory,
                    "total_memory": total_memory,
                    }
            result[i] = info
        return result


if __name__ == "__main__":
    gst = GpuStat()
    print(gst.gpu_stat())
    gst.stop()
    # method 2 使用gpustat库
    # from sparrow.cli.core import command, commandexists
    # print(command(["gpustat", "--no-color"]))



import time
import subprocess
import io
import csv
import collections


def run_job(job, num: int, interval="second"):
    import schedule

    if interval == "second":
        schedule.every(num).seconds.do(job)
    elif interval == "minute":
        schedule.every(num).minutes.do(job)
    elif interval == "hour":
        schedule.every(num).hours.do(job)
    elif interval == "day":
        schedule.every(num).days.do(job)
    while True:
        schedule.run_pending()
        time.sleep(1)


def commandexists(shellcommand):
    status, output = subprocess.getstatusoutput(shellcommand)
    exists = status == 0
    if not exists:
        print("Could not execute: {0}".format(shellcommand))
    return exists

def command(args):
    # subprocess.call(args)
    return subprocess.check_output(args).decode()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sparrow',
 'sparrow.algorithm',
 'sparrow.ann',
 'sparrow.api',
 'sparrow.cli',
 'sparrow.color',
 'sparrow.cv',
 'sparrow.datasets',
 'sparrow.debug',
 'sparrow.decorators',
 'sparrow.distance',
 'sparrow.doc',
 'sparrow.docker',
 'sparrow.functions',
 'sparrow.inspect',
 'sparrow.io',
 'sparrow.jupyter',
 'sparrow.jupyter.utils',
 'sparrow.log',
 'sparrow.math',
 'sparrow.math.sampling',
 'sparrow.multiprocess',
 'sparrow.multiprocess.mq',
 'sparrow.ner',
 'sparrow.net',
 'sparrow.nn',
 'sparrow.path',
 'sparrow.performance',
 'sparrow.progress_bar',
 'sparrow.proto',
 'sparrow.proto.python',
 'sparrow.proxy',
 'sparrow.string',
 'sparrow.table_ops',
 'sparrow.template',
 'sparrow.template.scaffold',
 'sparrow.template.scaffold.src',
 'sparrow.template.scaffold.src.project_name',
 'sparrow.template.scaffold.src.project_name.api',
 'sparrow.template.scaffold.src.project_name.api.routes',
 'sparrow.template.scaffold.src.tests',
 'sparrow.transform',
 'sparrow.trie',
 'sparrow.utils',
 'sparrow.web',
 'sparrow.widgets']

package_data = \
{'': ['*'],
 'sparrow.ann': ['milvus_db/*'],
 'sparrow.api': ['static/*'],
 'sparrow.log': ['conf/*'],
 'sparrow.proto': ['js/*'],
 'sparrow.template.scaffold.src': ['Examples/*', 'conf/*', 'docker/*'],
 'sparrow.web': ['.streamlit/*']}

install_requires = \
['Deprecated',
 'GitPython',
 'attrs>=22.1.0,<23.0.0',
 'chevron',
 'colour',
 'concurrent-log-handler',
 'fire',
 'gpustat>=1.0.0,<2.0.0',
 'numpy',
 'pandas>=1.5.0,<2.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'pretty_errors>=1.2.25,<2.0.0',
 'psutil>=5.9.2,<6.0.0',
 'pyyaml',
 'rich']

entry_points = \
{'console_scripts': ['spr = sparrow.__main__:main']}

setup_kwargs = {
    'name': 'sparrow-tool',
    'version': '0.8.1',
    'description': '',
    'long_description': '# sparrow_tool\n[![image](https://img.shields.io/badge/Pypi-0.8.1-green.svg)](https://pypi.org/project/sparrow_tool)\n[![image](https://img.shields.io/badge/python-3.6+-blue.svg)](https://www.python.org/)\n[![image](https://img.shields.io/badge/license-MIT-blue.svg)](LICENSE)\n[![image](https://img.shields.io/badge/author-kunyuan-orange.svg?style=flat-square&logo=appveyor)](https://github.com/beidongjiedeguang)\n\n\n-------------------------\n## Install\n```bash\npip install sparrow-tool\n# Or dev version\npip install sparrow-tool[dev]\n# Or\npip install -e .\n# Or\npip install -e .[dev]\n```\n\n\n## Usage\n\n### Safe logger in `multiprocessing`\n```python\nfrom sparrow.log import Logger\nimport numpy as np\nlogger = Logger(name=\'train-log\', log_dir=\'./logs\', )\nlogger.info("hello","numpy:",np.arange(10))\n\nlogger2 = Logger.get_logger(\'train-log\')\nprint(id(logger2) == id(logger))\n>>> True\n```\n\n### Multiprocessing SyncManager\n\nOpen server first:\n```bash\n$ sparrow start-server\n```\nThe defualt port `50001`.\n\n(Process1) productor:\n```python\nfrom sparrow.multiprocess.client import Client\nclient = Client(port=50001)\nclient.update_dict({\'a\': 1, \'b\': 2})\n```\n\n(Process2) consumer:\n```python\nfrom sparrow.multiprocess.client import Client\nclient = Client(port=50001)\nprint(client.get_dict_data())\n\n>>> {\'a\': 1, \'b\': 2}\n```\n\n### Common tools\n- **Kill process by port**\n    ```bash\n    $ sparrow kill {port}\n    ```\n\n- **pack & unpack**  \n    support archive format: "zip", "tar", "gztar", "bztar", or "xztar".\n    ```bash\n    sparrow pack pack_dir \n    sparrow unpack filename extract_dir\n    ```\n',
    'author': 'kunyuan',
    'author_email': 'beidongjiedeguang@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/beidongjiedeguang/sparrow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.0,<4',
}


setup(**setup_kwargs)



from __future__ import annotations

import os
import logging
from logging import handlers
from logging.config import dictConfig
from rich.logging import RichHandler
from pathlib import Path
from ..path import rel_to_abs
from ..io import yaml_load
from ..string.color_string import rgb_string, color_const


def setup_logging(config_path: str | Path = None,
                  log_path="./log/",
                  multi=False):
    """Setup logging from config file.

    Parameters
    ----------
    config_path : str|Path, default None
        logging config file(.yaml)
    log_path : str|Path
        Saved log dir
    multi : bool
        是否多进程中使用logger

    Examples
    --------
    >>> import logging
    >>> setup_logging()
    >>> logger = logging.getLogger("debug") # Optional: "debug", "info"
    """
    if config_path is None:
        config_path = rel_to_abs("./conf/logging.yaml")
    log_path = Path(log_path)
    if not log_path.exists():
        print(rgb_string(f"Not Found default log path, create dir: `{str(log_path)}`",
                         color=color_const.GOLD))
        log_path.mkdir()
    if os.path.exists(config_path):
        setattr(logging, "RichHandler", RichHandler)
        if multi:
            from concurrent_log_handler import ConcurrentRotatingFileHandler
            setattr(handlers, 'RotatingFileHandler', ConcurrentRotatingFileHandler)
        logging_config = yaml_load(config_path)
        origin_handlers = logging_config['handlers']
        for handler_name in origin_handlers:
            if handler_name.startswith('file'):
                handler_file = origin_handlers[handler_name]['filename']
                abs_handler_file = os.path.join(log_path, handler_file)
                origin_handlers[handler_name]['filename'] = abs_handler_file
        dictConfig(logging_config)
        print(rgb_string("Init logging by config success !", color=color_const.GREEN))
    else:
        raise SystemExit(f"logging config file {config_path} not found, exit main processing")



import os
from .io import yaml_load, yaml_dump, rm
from .string.ops import string_add


class VersionControl:
    def __init__(
            self,
            pkgname,
            pkgdir,
            version=None,
            filename="version-config.yaml",
    ):

        # self.config = None
        self._pkgname = pkgname
        self._config_path = os.path.join(pkgdir, filename)
        if version is None:
            try:
                self.get_config()
            except:
                print(f"{filename} was not exist, created now.")
                self.gen_config("0.0.0")
        else:
            self.gen_config(version)

    def gen_config(self, version="0.0.0"):
        config = {"name": self._pkgname, "version": version}
        self.config = config
        yaml_dump(self._config_path, config, rel_path=False)

    def get_config(self):
        config = yaml_load(self._config_path, rel_path=False)
        self.config = config
        return config

    def set_version(self, version):
        self.config["version"] = version

    def save_config(self):
        yaml_dump(self._config_path, self.config, rel_path=False)

    def update_version(self, version_step=1):
        self.config["version"] = string_add(self.config["version"], version_step)
        self.update_pyproject()
        yaml_dump(self._config_path, self.config, rel_path=False)

    def clean_config_file(self):
        os.remove(self._config_path)

    def update_pyproject(self):
        pyproject_path = "pyproject.toml"
        # import toml
        # pyproject = toml.load(pyproject_path)
        # pyproject['tool']['poetry']['version'] = self.config["version"]
        # pyproject['tool']['poetry']['version'] = 0.9
        # with open(pyproject_path, "w", encoding="utf8") as f:
        #     toml.dump(pyproject, f)
        with open(pyproject_path, "r", encoding="UTF-8") as fr:
            pyproject_list = fr.readlines()
        for idx, line in enumerate(pyproject_list[:10]):
            if "version" in line:
                new_line = f"""version = "{self.config['version']}"
"""
                pyproject_list[idx] = new_line
        with open(pyproject_path, 'w', encoding="UTF-8") as fw:
            fw.writelines(pyproject_list)

    def update_readme(
            self,
            readme_path="README.md",
            license="MIT",
            author="kunyuan",
            replace_flag=19 * "-",
    ):
        with open(readme_path, "r", encoding="UTF-8") as fr:
            readme = fr.read()
        replace_begin = f"""\
# {self._pkgname}
[![image](https://img.shields.io/badge/Pypi-{self.config['version']}-green.svg)](https://pypi.org/project/{self._pkgname})
[![image](https://img.shields.io/badge/python-3.6+-blue.svg)](https://www.python.org/)
[![image](https://img.shields.io/badge/license-{license}-blue.svg)](LICENSE)
[![image](https://img.shields.io/badge/author-{author}-orange.svg?style=flat-square&logo=appveyor)](https://github.com/beidongjiedeguang)


"""
        readme_list = readme.split(replace_flag)
        readme_list[0] = replace_begin
        new_readme = replace_flag.join(readme_list)

        with open(readme_path, "w", encoding="UTF-8") as fo:
            fo.write(new_readme)

    def upload_pypi(self):
        pkgname = self._pkgname
        rm("build", "dist", "eggs", f"{pkgname}.egg-info")
        os.system("python setup.py sdist bdist_wheel")
        os.system("twine upload dist/*")
        rm("build", "dist", "eggs", f"{pkgname}.egg-info")

    def install(self):
        pkgname = self._pkgname
        rm("build", "dist", "eggs", f"{pkgname}.egg-info")
        os.system(f"pip uninstall {pkgname} -y && python setup.py install")
        rm("build", "dist", "eggs", f"{pkgname}.egg-info")

    @staticmethod
    def build():
        os.system(f"pythom -m build")



import streamlit as st
from io import StringIO
from queue import Queue
from collections import deque
from dataclasses import dataclass

if 'file_state' not in st.session_state:
    @dataclass
    class FileState:
        data: deque
        name: deque
        type: deque


    st.session_state['file_state'] = {"data": deque(maxlen=2), "name": deque(maxlen=2), "type": deque(maxlen=2)}

file_state = st.session_state['file_state']

uploaded_files = st.file_uploader("Choose file", accept_multiple_files=True)
print(uploaded_files)
if uploaded_files:
    # To read file as bytes:
    for uploaded_file in uploaded_files:
        file_state['data'].append(uploaded_file.getvalue())
        file_state['name'].append(uploaded_file.type)
        file_state['type'].append(uploaded_file.type)

        if "image/" in uploaded_file.type:
            st.image(uploaded_file)
        elif 'video' in uploaded_file.type:
            st.video(uploaded_file)

    # To convert to a string based IO:
    # stringio = StringIO(uploaded_file.getvalue().decode("utf-8"))
    # st.write(stringio)
    #
    # # To read file as string:
    # string_data = stringio.read()
    # st.write(string_data)
    #
    # # Can be used wherever a "file-like" object is accepted:
    # dataframe = pd.read_csv(uploaded_file)
    # st.write(dataframe)

if file_state['name']:
    st.download_button('Download file', file_state['data'][-1],
                       mime=file_state['type'][-1],
                       # file_name=file_state['name'][-1]
                       )



import numpy as np
from .bezier import bezier


def linear(t):
    return t


def smooth(t):
    # Zero first and second derivatives at t=0 and t=1.
    # Equivalent to bezier([0, 0, 0, 1, 1, 1])
    s = 1 - t
    return (t**3) * (10 * s * s + 5 * s * t + t * t)


def rush_into(t):
    return 2 * smooth(0.5 * t)


def rush_from(t):
    return 2 * smooth(0.5 * (t + 1)) - 1


def slow_into(t):
    return np.sqrt(1 - (1 - t) * (1 - t))


def double_smooth(t):
    if t < 0.5:
        return 0.5 * smooth(2 * t)
    else:
        return 0.5 * (1 + smooth(2 * t - 1))


def there_and_back(t):
    new_t = 2 * t if t < 0.5 else 2 * (1 - t)
    return smooth(new_t)


def there_and_back_with_pause(t, pause_ratio=1.0 / 3):
    a = 1.0 / pause_ratio
    if t < 0.5 - pause_ratio / 2:
        return smooth(a * t)
    elif t < 0.5 + pause_ratio / 2:
        return 1
    else:
        return smooth(a - a * t)


def running_start(t, pull_factor=-0.5):
    return bezier([0, 0, pull_factor, pull_factor, 1, 1, 1])(t)


def not_quite_there(func=smooth, proportion=0.7):
    def result(t):
        return proportion * func(t)

    return result


def wiggle(t, wiggles=2):
    return there_and_back(t) * np.sin(wiggles * np.pi * t)


def squish_rate_func(func, a=0.4, b=0.6):
    def result(t):
        if a == b:
            return a
        elif t < a:
            return func(0)
        elif t > b:
            return func(1)
        else:
            return func((t - a) / (b - a))

    return result


# Stylistically, should this take parameters (with default values)?
# Ultimately, the functionality is entirely subsumed by squish_rate_func,
# but it may be useful to have a nice name for with nice default params for
# "lingering", different from squish_rate_func's default params


def lingering(t):
    return squish_rate_func(lambda t: t, 0, 0.8)(t)


def exponential_decay(t, half_life=0.1):
    # The half-life should be rather small to minimize
    # the cut-off error at the end
    return 1 - np.exp(-t / half_life)



import numpy as np


def get_hilbert_1d_array(image: np.ndarray):
    from hilbert import decode, encode  # $ pip install numpy-hilbert-curve
    num_dims = 2
    num_bits = np.log2(image.shape[0] * image.shape[1]) / num_dims
    num_bits = int(num_bits)
    max_hil = 2 ** (num_bits * num_dims)
    hilberts = np.arange(max_hil)
    locs = decode(hilberts, num_dims, num_bits)

    image1d = []
    for coord in locs:
        image1d.append(image[coord[0], coord[1]])
    return np.array(image1d)[None, ...]
